/*
    Copyright (C) 2011 Nokia Corporation and/or its subsidiary(-ies).

    Contact: Ruslan Mstoi <ruslan.mstoi@nokia.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301  USA.
*/

#include <QtCore>
#include <QtGui>
#include <QApplication>

#include <iostream>

#include "qgraphicstermwidget.h"

using std::cout;

struct Options
{
    Options():
        cmd(),    // empty
        cmdArgs() // empty
    {}

    QString cmd;         // command specified with the -e option
    QStringList cmdArgs; // arguments to pass to the cmd
};

void printHelp(QString appName)
{
    cout << "Usage: " << qPrintable(appName) << " [options] [args]\n\n";
    cout << "Terminal emulator\n\n";
    cout << "Options:\n";
    cout << "  -h, --help\t Show this help message and exit\n";
    cout << "  -e CMD\t Command to execute, options and arguments after "
        "the CMD will be passed to the CMD\n\n";
}

/**
 * Parses arguments string list, like one returned by
 * QCoreApplication::arguments()
 */
void parseArgsList(QStringList args, Options &options)
{
    for (int i = 0; i < args.size(); ++i) {
        QString s = args.at(i);
        
        if (s == "-h" || s.startsWith("-help") || s.startsWith("--help")) {
            printHelp(QFileInfo(args.at(0)).fileName());
            // let MComponentDataPrivate::parseArguments print its help and exit
            //return;
            exit(0);
        }

        else if (s == "-e") {
            if (i == args.size() - 1) { // error: -e is last, cmd not specified
                cout << "Please specify command to execute!\n";
                exit(1);
            }

            QString cmd = args.at(i + 1);
            if (cmd.startsWith(QLatin1String("./"))) {
                cmd = QDir::currentPath() + cmd.mid(1);
            }

            QStringList cmdArgs = args.mid(i + 1); // cmd is 1st element

            options.cmd = cmd;
            options.cmdArgs = cmdArgs;

            // -e is last option, everything after it is passed to the command
            break;
        }
    }
}

int main(int argc, char *argv[])
{
    int w = 800, h = 600;
    QApplication app(argc, argv);

    Options options;
    parseArgsList(app.arguments(), options);

    QGraphicsScene scene;
    QGraphicsView view(&scene);
    view.setBackgroundBrush(Qt::blue);

    QGraphicsTermWidget *console = new QGraphicsTermWidget(false, 0);

    if (!options.cmd.isEmpty()) {
        console->setShellProgram(options.cmd);
        console->setArgs(options.cmdArgs);
    }
    console->startShellProgram();

    QFont font = QApplication::font();
    font.setFamily("Monospace");
    font.setPointSize(12);
    console->setTerminalFont(font);

    console->resize(w, h);
    //console->setColorScheme(COLOR_SCHEME_BLACK_ON_LIGHT_YELLOW);
    scene.addItem(console);
    view.resize(w + 20, h + 20);

    QObject::connect(console, SIGNAL(finished()), &app, SLOT(quit()));

    view.show();

    return app.exec();
} 
