/*
    This file is part of meego-terminal

    Copyright (C) 2011 Nokia Corporation and/or its subsidiary(-ies).

    Contact: Ruslan Mstoi <ruslan.mstoi@nokia.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301  USA.
*/

#include <MPannableViewport>
#include <MInputMethodState>
#include <MPositionIndicator>
#include <MMessageBox>
#include <MBanner>

#include "MTermWidget.h"
#include "MTerminalDisplay.h"
#include "Session.h"
#include "ScreenWindow.h"

// multiplier that converts display lines into viewport values, purpose is to
// make panning easier to use (i.e. not too sensitive)
#define PANM 10 

/**
 * [1] Display cannot be set as viewports widget because it handles history
 * itself, so it does not render all of its contents as viewport expects. Also,
 * viewport does not support abstract ranges but range is in actual
 * coordinates. So, fake widget is used to trick the viewport to handle the pan
 * gestures. This class does the conversion from viewport coordinates into
 * terminal lines.
 */
MTermWidget::MTermWidget(int startnow, QGraphicsWidget *parent):
    QGraphicsTermWidget(parent), // do custom session and display creation
    m_lastSetCurrentLine(0),
    m_colorScheme(2) // green on black
{
    construct(startnow);

    setAutoFillBackground(true); // see MTermWidget::setColorScheme [1]

    m_viewport = new MPannableViewport(this);
    m_viewport->setAutoRange(false);

    // note fake widget, see [1]
    m_viewport->setWidget(new QGraphicsWidget(this));

    connect(m_terminalDisplay, SIGNAL(scrollChanged(int, int)),
            this, SLOT(updateViewport(int, int)));

    connect(m_viewport, SIGNAL(positionChanged(const QPointF &)),
            this, SLOT(viewportPositionChanged(const QPointF &)));

    connect(MInputMethodState::instance(),
            SIGNAL(inputMethodAreaChanged(QRect)),
            this, SLOT(onInputMethodAreaChanged(QRect)),
            Qt::UniqueConnection);

    connect(m_terminalDisplay, SIGNAL(changedFontMetricSignal(int, int)),
            this, SLOT(displayFontChanged()));

    grabGesture(Qt::SwipeGesture);
    grabGesture(Qt::PinchGesture);

    readSettings();
}

/**
 * Creates session and display and does initialization. Basically does the same
 * as the non-default base constructor, but creates MTerminalDisplay instead of
 * TerminalDisplay.
 */
void MTermWidget::construct(int startnow)
{
    m_session = createSession();
    m_terminalDisplay = createTerminalDisplay(m_session);

    init();

    if (startnow && m_session) {
        m_session->run();
    }

    // set focus policy so that display gets focus, default focus policy is
    // Qt::NoFocus which means ItemIsFocusable flag is not set
     this->setFocusPolicy( m_terminalDisplay->focusPolicy() ); 

    this->setFocus( Qt::OtherFocusReason );
    m_terminalDisplay->resize(this->size());
    
    this->setFocusProxy(m_terminalDisplay);
}

MTermWidget::~MTermWidget()
{
    writeSettings(); // could do it also in sessionFinished
}

TerminalDisplay *MTermWidget::createTerminalDisplay(Session *session)
{
    TerminalDisplay* display = new MTerminalDisplay(this);

    display->setBellMode(TerminalDisplay::NotifyBell);
    display->setTerminalSizeHint(true);
    display->setTripleClickMode(TerminalDisplay::SelectWholeLine);
    display->setTerminalSizeStartup(true);

    display->setRandomSeed(session->sessionId() * 31);
    
    return display;
}

void MTermWidget::resizeEvent(QGraphicsSceneResizeEvent *event)
{
    QRect imArea = MInputMethodState::instance()->inputMethodArea();

    if (imArea.isEmpty()) { // vkb is not up
        QGraphicsTermWidget::resizeEvent(event);
        m_viewport->resize(this->size());
    }
    // vkb is up
    else {
        int newHeight = 0; // = where vkb starts - status bar
        
        if (imArea.x() == 0) { // landscape
            // Bug in the framework? in landscape mode status bar is hidden and
            // terminal is shown in fullscreen mode, but scenePos() still
            // returns QPointF(0, 36) but it should be (0, 0). So
            // scenePos().y() is not used for now.
            //newHeight = imArea.y() - scenePos().y();
            newHeight = imArea.y();
        }

        else if (imArea.y() == 0) { // portrait
            newHeight = imArea.x() - scenePos().x();
        }

        if (newHeight > 0) {  // resize to occupy area not taken  by vkb
            m_terminalDisplay->resize(size().width(), newHeight);
            m_viewport->resize(size().width(), newHeight);
        }
    }
}

/**
 * At least when coming to portrait mode this method is called after
 * resizeEvent, so trigger resizeEvent. Parameter is null, it is not used
 * anyway.
 */
void MTermWidget::onInputMethodAreaChanged(const QRect &imArea)
{
    Q_UNUSED(imArea);
    resizeEvent(0);
}

void MTermWidget::updateViewport(int currentLine, int maxLines)
{
    QRectF newRange = m_viewport->range();
    newRange.setHeight(maxLines * PANM);

    QPointF newPos(m_viewport->position().x(), currentLine * PANM);

    // avoid rounding error when called due to viewportPositionChanged
    if (m_lastSetCurrentLine == currentLine && maxLines) {
        newPos.setY(m_viewport->position().y());
    }

    if (newRange != m_viewport->range()) {
        m_viewport->setRange(newRange);
    }

    if (newPos != m_viewport->position()) {
        m_viewport->setPosition(newPos);
    }
}

void MTermWidget::viewportPositionChanged(const QPointF &position)
{
    static qreal prevPosY = 0;
    qreal posY = position.y();

    if (posY <  0) {// spring action this
        posY = 0;
    }

    if (posY > m_viewport->range().height()) {// spring action this
        posY = m_viewport->range().height();
    }
    
    if (prevPosY == posY) {
        return;
    }

    prevPosY = posY;
    m_lastSetCurrentLine = posY / PANM;

    const bool atEndOfOutput = (posY == m_viewport->range().height());
    m_terminalDisplay->viewportPositionChanged(m_lastSetCurrentLine,
                                               atEndOfOutput);
}

bool MTermWidget::event(QEvent *event)
{
    if (event->type() == QEvent::Gesture)
        return gestureEvent(static_cast<QGestureEvent*>(event));
    return QGraphicsTermWidget::event(event);
}

bool MTermWidget::gestureEvent(QGestureEvent *event)
{
    if (QGesture *swipe = event->gesture(Qt::SwipeGesture))
         swipeTriggered(static_cast<QSwipeGesture *>(swipe));
    if (QGesture *pinch = event->gesture(Qt::PinchGesture))
        pinchTriggered(static_cast<QPinchGesture *>(pinch));
    
    return true;
}

/**
 * The complexity in handling swipe directions is because QSwipeGesture does
 * not understand orientation change. For example, after a change from
 * landscape to portrait horizontal swipes are reported for vertical ones and
 * vice versa.
 */
void MTermWidget::swipeTriggered(QSwipeGesture *gesture)
{
    if (gesture->state() != Qt::GestureFinished) {
        return;
    }

    static const int min = COLOR_SCHEME_WHITE_ON_BLACK; // 1
    static const int max = COLOR_SCHEME_BLACK_ON_LIGHT_YELLOW; // 3

    // defaults assume landscape orientation
    QSwipeGesture::SwipeDirection otherDirection =
        gesture->verticalDirection();
    QSwipeGesture::SwipeDirection currentDirection =
        gesture->horizontalDirection();
    QSwipeGesture::SwipeDirection prevDirection = QSwipeGesture::Left;
    QSwipeGesture::SwipeDirection nextDirection = QSwipeGesture::Right;

    // portrait (at the time of writing this QSystemDisplayInfo.orientation did
    // not work, it always returned QSystemDisplayInfo::Landscape)
    if (size().width() == 480) {
        otherDirection = gesture->horizontalDirection();
        currentDirection = gesture->verticalDirection();
        prevDirection = QSwipeGesture::Up;
        nextDirection = QSwipeGesture::Down;
    }

    // accept only clean swipes in the needed direction, i.e. no simultaneous
    // accidental swipes in both directions
    if (otherDirection != QSwipeGesture::NoDirection) {
        return;
    }

    int origColorScheme = m_colorScheme;

    if (currentDirection == prevDirection) { // select previous scheme
        m_colorScheme -= 1;
        if (m_colorScheme < min) {
            m_colorScheme = max;
        }
    }
    else if (currentDirection == nextDirection) { // select next scheme
        m_colorScheme += 1;
        if (m_colorScheme > max) {
            m_colorScheme = min;
        }
    }

    if (origColorScheme != m_colorScheme)
        setColorScheme(m_colorScheme);
}

void MTermWidget::pinchTriggered(QPinchGesture *gesture)
{
    QPinchGesture::ChangeFlags changeFlags = gesture->changeFlags();

    // initially scaleFactor is 1.0
    static const qreal scaleFactorInit = 1.0;
    static qreal lastTotalScaleFactor = scaleFactorInit;
        
    if (changeFlags & QPinchGesture::ScaleFactorChanged) {
        qreal totalScaleFactor = gesture->totalScaleFactor();
        qreal diff = totalScaleFactor - lastTotalScaleFactor;

        if (diff > .7) {
            m_terminalDisplay->increaseTextSize();
            lastTotalScaleFactor = totalScaleFactor;
        }
        if (diff < -.3) {
            m_terminalDisplay->decreaseTextSize();
            lastTotalScaleFactor = totalScaleFactor;
        }
    }
    if (gesture->state() == Qt::GestureFinished) {
        lastTotalScaleFactor = scaleFactorInit;
    }
}

void MTermWidget::displayFontChanged()
{
    QFont font = m_terminalDisplay->getVTFont();

    showBanner(QString("Font size is now %1 ").arg(font.pointSize()));
}

void MTermWidget::showBanner(const QString &title)
{
    MBanner *banner = new MBanner();
    // InformationBanner takes up too much space
    // banner->setStyleName("InformationBanner"); 
    banner->setStyleName("ShortEventBanner");
    banner->setTitle(title);
    banner->appear(scene(), MSceneWindow::DestroyWhenDone);
}

void MTermWidget::readSettings()
{
    QSettings settings("Nokia", "meego-terminal");

    int fontPointSize = settings.value("fontPointSize", 12).toInt();
    m_colorScheme = settings.value("colorScheme",
                                   COLOR_SCHEME_GREEN_ON_BLACK).toInt();
    
    // FIXME: because of lexical comparisson of versions, it will not work if
    // formats are different (i.e.) dots are in different places
    QString currentVersion = "0.1.7"; // current version of the terminal

    // last version for which "what's new" dialog was shown
    QString lastWhatsNewVersion =
        settings.value("lastWhatsNewVersion").toString();

    QFont font = QApplication::font(); 
    font.setFamily("Monospace");
    font.setPointSize(fontPointSize);
    setTerminalFont(font);

    setColorScheme(m_colorScheme);

    if (lastWhatsNewVersion < currentVersion) {
        QString title = QString("What's new in MeeGo Terminal %1?"
                                ).arg(currentVersion);

        QString text = "Pinch to change font size\n\n"
            "Swipe horizontally to change color scheme (experimental feature)";

        MMessageBox *messageBox = new MMessageBox(title, text);
        messageBox->appear(MSceneWindow::DestroyWhenDone);

        settings.setValue("lastWhatsNewVersion", currentVersion); // shown only once
    }
}

void MTermWidget::writeSettings()
{
    QSettings settings("Nokia", "meego-terminal");

    settings.setValue("fontPointSize", m_terminalDisplay->getVTFont().pointSize());
    settings.setValue("colorScheme", m_colorScheme);
}

/**
 * [1] Background color is set to the one in the scheme of the terminal
 * display. This is to avoid flicker when display is resized due to vkb
 * show/hide events. The display is resized in resizeEvent, but vkb is still
 * animating, though it reports full size, so for a while white default
 * background of this widget becomes visible: hence the flicker.
 */
void MTermWidget::setColorScheme(int scheme)
{
    static QString scheme2text[] = { "UNKNOWN",
                                     "White on Black",
                                     "Green on Black",
                                     "Black on Light Yellow" };

    QGraphicsTermWidget::setColorScheme(scheme);
    m_colorScheme = scheme;

    QColor bgColor = m_terminalDisplay->colorTable()[DEFAULT_BACK_COLOR].color;

    QPalette p = palette(); // see [1]
    p.setColor(QPalette::Window, bgColor);
    setPalette(p);

    // use white indicator if background is black
    QString style = "CommonPositionIndicatorInverted";
    if (m_colorScheme == COLOR_SCHEME_BLACK_ON_LIGHT_YELLOW) {
	style = "CommonPositionIndicator"; // default style
    }
    if (style != m_viewport->positionIndicator()->styleName())
	m_viewport->positionIndicator()->setStyleName(style);

    showBanner(QString("Color scheme is now '%1'").arg(scheme2text[scheme]));
}

/**
 * Overridden to check if the session changes the program in its run method. If
 * it does, then the user has specified invalid program name with the -e option
 * and the session has changed it to a shell program. In that case, shows a
 * message box explaining the case.
 */
void MTermWidget::startShellProgram()
{
    if ( m_session->isRunning() )
        return;

    QString oldProgram = m_session->program();
    QGraphicsTermWidget::startShellProgram();
    QString newProgram = m_session->program();

    if (oldProgram != newProgram) {
        QString title = QString("Warning!");
        QString text = QString("Could not find '%1', starting '%2' instead."
                               ).arg(oldProgram).arg(newProgram);

        MMessageBox *messageBox = new MMessageBox(title, text);
        messageBox->setIconId("icon-l-error");
        messageBox->appear(MSceneWindow::DestroyWhenDone);
    }
}

