/*
    This file is part of meego-terminal

    Copyright (C) 2011 Nokia Corporation and/or its subsidiary(-ies).

    Contact: Ruslan Mstoi <ruslan.mstoi@nokia.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301  USA.
*/

#include <QApplication>
#include <QGraphicsView>
#include <QX11Info>
#include <MInputMethodState>

#include "MTerminalDisplay.h"

#include <X11/Xutil.h>

// KeyPress in QEvent::KeyPress will be expanded by preprocessor to 2 because
// of definition of KeyPress in X11 headers, which would result in weird
// compiler errors as "expected unqualified-id before numeric constant". So
// solution from qt_x11_p.h:
// rename a couple of X defines to get rid of name clashes
// resolve the conflict between X11's FocusIn and QEvent::FocusIn
enum {
    XFocusOut = FocusOut,
    XFocusIn = FocusIn,
    XKeyPress = KeyPress,
    XKeyRelease = KeyRelease,
    XNone = None,
    XRevertToParent = RevertToParent,
    XGrayScale = GrayScale,
    XCursorShape = CursorShape
};

#undef FocusOut
#undef FocusIn
#undef KeyPress
#undef KeyRelease
#undef None
#undef RevertToParent
#undef GrayScale
#undef CursorShape

MTerminalDisplay::MTerminalDisplay(QGraphicsWidget *parent):
    TerminalDisplay(parent),
    m_registeredToolbarId(0),
    m_ctrlButtonIsChecked(false)
{
    QString name = "/usr/share/meego-terminal/meego-terminal-toolbar.xml";
    m_registeredToolbarId = MInputMethodState::instance()->registerToolbar(name);
}

MTerminalDisplay::~MTerminalDisplay()
{
    MInputMethodState::instance()->unregisterToolbar(m_registeredToolbarId);
}

void MTerminalDisplay::keyPressEvent( QKeyEvent* event )
{
    // Since right now there is no way to read checked state of an input method
    // toolbar button, Control toggle button on the input method toolbar sends
    // key sequence: Ctrl+Shift+Alt+X
    if (event->key() == Qt::Key_X &&
        (event->modifiers() &
         (Qt::ControlModifier | Qt::ShiftModifier | Qt::AltModifier))) {

        m_ctrlButtonIsChecked = !m_ctrlButtonIsChecked;
        event->accept();
        return;
    }

    TerminalDisplay::keyPressEvent(event);
}

/**
 * Takes plain text (which is assumed to be single character representing
 * single key press as provided by the input method) and then controllifies
 * it. That is, returns text that XLookupString returns for that character and
 * control modifier. This is all done to provide the emulation with the proper
 * text() of QKeyEvent when Ctrl toggle button of the input method toolbar is
 * checked.
 *
 */
QString MTerminalDisplay::getCtrlifiedText(const QString& plainText)
{
    QString text;

    // too bad QGraphicsWidget does not have x11Info
    QGraphicsView *focusedGraphicsView =
        dynamic_cast<QGraphicsView *>(QApplication::focusWidget());

    if (!focusedGraphicsView) {
        return plainText;
    }

    QX11Info x11Info = focusedGraphicsView->x11Info();

    // could use XStringToKeysym() to get keysym, but for latin letters
    // keysym == ascii
    KeySym keySym = plainText.toAscii()[0];
    KeyCode keyCode = XKeysymToKeycode(x11Info.display(), keySym);

    if (!keyCode) {
        return plainText;
    }

    XKeyEvent xev; // fake event with the pressed key code and control modifier
    xev.type = XKeyPress;
    xev.serial = 0;
    xev.send_event = False;
    xev.display = x11Info.display();
    xev.window = XNone;
    xev.root = XNone;
    xev.subwindow = XNone;
    xev.time = 0;
    xev.x = 0;
    xev.y = 0;
    xev.x_root = 0;
    xev.y_root = 0;
    xev.state = ControlMask;
    xev.keycode = keyCode;
    xev.same_screen = 1;

    // from QKeyMapperPrivate::translateKeyEventInternal:
    // some XmbLookupString implementations don't return buffer overflow correctly,
    // so we increase the input buffer to allow for long strings...
    // 256 chars * 2 bytes + 1 null-term == 513 bytes
    QByteArray chars;
    chars.resize(513);
    int count = 0;

    count = XLookupString(&xev, chars.data(), chars.size(), &keySym, 0);

    if (!count) {
        return plainText;
    }

    // if stops working consider using mapper as in Qt's translateKeySym
    text = QString::fromLatin1(chars, count);

    return text;
}

void MTerminalDisplay::inputMethodEvent(QInputMethodEvent* event)
{
    // same stuff as TerminalDisplay::inputMethodEvent with Ctrl modifier applied
    if (m_ctrlButtonIsChecked) {

        QString text = getCtrlifiedText(event->commitString());

        // ControlModifier redundant, emulation only uses text
        QKeyEvent keyEvent(QEvent::KeyPress,
                           0,
                           Qt::ControlModifier,
                           text);
        emit keyPressedSignal(&keyEvent);

        // Graphics View Framework: this slows down text entry with hw keyboard (cursor
        // movement becomes lazy)
#if 0 
        _inputMethodData.preeditString = event->preeditString();
        update(preeditRect() | _inputMethodData.previousPreeditRect);
#endif
    
        event->accept();
    }

    else {
        TerminalDisplay::inputMethodEvent(event);
    }
}

QVariant MTerminalDisplay::inputMethodQuery(Qt::InputMethodQuery query) const
{
    switch ((int)query) {

    case M::InputMethodAttributeExtensionIdQuery:
        return QVariant(m_registeredToolbarId);
        
    default:
        return TerminalDisplay::inputMethodQuery(query);
    }
}
